#include <stdio.h> /* for snprintf */
#include <string.h>
#include <assert.h>
#include <allegro.h>
#include "config.h"
#include "ctimer.h"
#include "crandom.h"
#include "game.h"
#include "field.h"
#include "images.h"
#include "font.h"
#include "main.h"
#include "colors.h"
#include "sound.h"

typedef struct
{
	int y, x;
	int hdir; /* -1 left, 0 none, 1 right */
	int vdir; /* -1 up,   0 none, 1 down  */
	int turn; /*          0 left, 1 right */
	int inuse;
	int crashed;
	int steps; /* how many steps since last move */

	/* Need this to teleport only once at a time. */
	int justteleported;

	/* And don't reflect forever, either. */
	int justreflected;

	/* You can't fire multiple times from the same spot. */
	int canfire;

	/* For restarting the level. */
	int spawny, spawnx;
	int gameovercount; /* When this hits 0, you restart. */
} train_t;
static train_t train;

#define MAXMONSTERS 256
typedef struct
{
	int y, x;
	int hdir;
	int vdir;
	int alive; /* hack: set to -1 to indicate collapsing */
	int steps;
} monster_t;
static monster_t monsters[MAXMONSTERS];
static int nummonsters;
static int numlivingmonsters;
#define TIMELIMIT 60 /* Low, so you need monster-kill bonuses badly */
static int timetics; /* time before you lose */

#define MAXBULLETS 256
typedef struct
{
	int y, x;
	int hdir;
	int vdir;
	int inuse;
	int steps;
} bullet_t;
static bullet_t bullets[MAXBULLETS];
static int numbullets;

/* Used to do things like "move the foo every 'bar' tics" */
static int tics;

#define LEFT -1
#define UP -1
#define NONE 0
#define DOWN 1
#define RIGHT 1

#define clipfunc(name, constant)		\
static int name(int n)				\
{						\
	while (n < 1) n += constant;		\
	while (n > constant) n -= constant;	\
	return n;				\
}
clipfunc(clipx, FIELD_WIDTH);
clipfunc(clipy, FIELD_HEIGHT);

static void spawn_train(int y, int x)
{
	train.y = train.spawny = y;
	train.x = train.spawnx = x;
	train.hdir = 1; /* east */
	train.vdir = 1; /* south */
	train.turn = 1; /* right */
	train.inuse = 1;
	train.crashed = 0;
	train.steps = 0;
	train.justteleported = 0;
	train.justreflected = 0;
	train.canfire = 1;
	train.gameovercount = 0;
}

static void setup_train(void)
{
	int y, x, spawned = 0;
	for (y = 1; y < FIELD_HEIGHT; y++)
		for (x = 1; x < FIELD_WIDTH; x++)
			if (field_get(y, x) == TRAIN_START)
			{
				if (!spawned) spawn_train(y, x);
				spawned = 1;
				field_set(y, x, TRAIN_REFLECT);
			}
}

static int monsterdsty(int mt)
{
	return clipy(monsters[mt].y + monsters[mt].vdir);
}
static int monsterdstx(int mt)
{
	return clipx(monsters[mt].x + monsters[mt].hdir);
}

/*
 * Tests if a monster other than exempt is in, or moving into, the spot.
 * If not, returns -1. If so, returns the monster's number.
 */
static int monsterinspot(int y, int x, int exempt)
{
	int i;
	for (i = 0; i < nummonsters; i++)
	{
		if (!monsters[i].alive || i == exempt) continue;
		if (monsters[i].x == x && monsters[i].y == y) return i;
		if (monsters[i].steps >= 0
			&& monsterdstx(i) == x && monsterdsty(i) == y)
		{
			return i;
		}
	}
	return -1;
}

static void spawn_bullet(int y, int x, int vdir, int hdir)
{
	if (numbullets == MAXBULLETS) abort_with_error("Too many monsters");
	bullets[numbullets].y = y;
	bullets[numbullets].x = x;
	bullets[numbullets].hdir = hdir;
	bullets[numbullets].vdir = vdir;
	bullets[numbullets].inuse = 1;
	bullets[numbullets].steps = 0;
	numbullets++;
}

static void remove_bullet(int num)
{
	bullets[num].inuse = 0;
	while (numbullets > 0 && bullets[numbullets-1].inuse == 0)
		numbullets--;
}

#define MONSTERKILLTIMEBONUS 12
#define MONCOLLAPSESTEPS 3

static void newlevel(void); /* Defined below, but needed here. */

static void killmonster(int mon)
{
	monsters[mon].alive = -MONCOLLAPSESTEPS; /* hack: collapse */
	numlivingmonsters--;
	sound_play(SMP_KILL);
	timetics += MONSTERKILLTIMEBONUS * TICS;
}

static int possiblyremovebullet(int bt)
{
	int mon;
	if (field_get(bullets[bt].y, bullets[bt].x) != EMPTY)
	{
		remove_bullet(bt);
		sound_play(SMP_DISINTEGRATE);
		return 1;
	}
	if ((mon = monsterinspot(bullets[bt].y, bullets[bt].x, -1)) != -1)
	{
		remove_bullet(bt);
		killmonster(mon);
		return 1;
	}
	return 0;
}

static void spawn_monster(int y, int x, int vdir, int hdir)
{
	if (nummonsters == MAXMONSTERS) abort_with_error("Too many monsters");
	monsters[nummonsters].y = y;
	monsters[nummonsters].x = x;
	monsters[nummonsters].hdir = hdir;
	monsters[nummonsters].vdir = vdir;
	monsters[nummonsters].alive = 1;
	monsters[nummonsters].steps = -1;
	nummonsters++;
	numlivingmonsters++;
}

/* Spawn num monsters in random empty spots. */
static void setup_monsters(int num)
{
	int y, x;
	nummonsters = 0;
	numlivingmonsters = 0;
	while (nummonsters < num)
	{
		y = rnd(FIELD_HEIGHT) + 1;
		x = rnd(FIELD_WIDTH)  + 1;
		if (field_get(y, x) == EMPTY
			&& monsterinspot(y, x, -1) == -1)
		{
			int movedir = rnd(4), hdir = NONE, vdir = NONE;
			if (movedir == 0) hdir = LEFT;
			else if (movedir == 1) hdir = RIGHT;
			else if (movedir == 2) vdir = UP;
			else if (movedir == 3) vdir = DOWN;
			spawn_monster(y, x, vdir, hdir);
		}
	}
}

int gameover = 0;

/* Restart a level after having run out of time. */
static void restartlevel(void)
{
	spawn_train(train.spawny, train.spawnx);
	setup_monsters(16);
	numbullets = 0;
	tics = 0;
	timetics = TIMELIMIT * TICS;
	sound_start();
}

static int titlescreen = 1;

static void newlevel(void)
{
	field_load();
	setup_train();
	setup_monsters(16);
	numbullets = 0;
	tics = 0;
	timetics = TIMELIMIT * TICS;
	titlescreen = 0;

	setup_colors();
	sound_nextmusic();
}

void game_init(void)
{
	load_images();
}

/************************************************************/

/*
 * I specifically chose STEPS_PER_MOVE to be the same as FONT_WIDTH
 * and FONT_HEIGHT, in font.h. Now it's different, but FONT_WIDTH and
 * FONT_HEIGHT are a multiple, so it's good enough I guess.
 */

#define TICS_PER_STEP 3 /* Train makes a "step" once every this many tics */
#define TICS_PER_MONSTER_STEP 6
#define TICS_PER_BULLET_STEP 3
#define STEPS_PER_MOVE 8 /* Number of steps for a full move */
#define MONSTER_WAIT 4      /* Monster waits this many steps out */
#define MONSTER_TURN_WAIT 2 /* Monster waits this many steps after turning */
#define TURN_LEFT 0
#define TURN_RIGHT 1
#define CRASH do { train.crashed = 1; return; } while(0)
enum { MOV_HORIZ, MOV_VERT, MOV_ASCEND, MOV_DESCEND };

static int movtype(int vdir, int hdir)
{
	if (vdir == NONE) return MOV_HORIZ;
	if (hdir == NONE) return MOV_VERT;
	if (vdir == hdir) return MOV_DESCEND;
	return MOV_ASCEND;
}

static void doturn(int *vdir, int *hdir, int turn)
{
	if (turn == TURN_LEFT)
	{
		if (*vdir == NONE && *hdir == LEFT) *vdir = DOWN;
		else if (*vdir == DOWN && *hdir == LEFT) *hdir = NONE;
		else if (*vdir == DOWN && *hdir == NONE) *hdir = RIGHT;
		else if (*vdir == DOWN && *hdir == RIGHT) *vdir = NONE;
		else if (*vdir == NONE && *hdir == RIGHT) *vdir = UP;
		else if (*vdir == UP && *hdir == RIGHT) *hdir = NONE;
		else if (*vdir == UP && *hdir == NONE) *hdir = LEFT;
		else if (*vdir == UP && *hdir == LEFT) *vdir = NONE;
	}
	else
	{
		if (*vdir == NONE && *hdir == LEFT) *vdir = UP;
		else if (*vdir == UP && *hdir == LEFT) *hdir = NONE;
		else if (*vdir == UP && *hdir == NONE) *hdir = RIGHT;
		else if (*vdir == UP && *hdir == RIGHT) *vdir = NONE;
		else if (*vdir == NONE && *hdir == RIGHT) *vdir = DOWN;
		else if (*vdir == DOWN && *hdir == RIGHT) *hdir = NONE;
		else if (*vdir == DOWN && *hdir == NONE) *hdir = LEFT;
		else if (*vdir == DOWN && *hdir == LEFT) *vdir = NONE;
	}
}

/*
 * Look for a secondary connection for the train, supposing it
 * turns in the direction turn. Return 1 if one is found.
 */
static int secondary(int turn)
{
	int dstx, dsty;
	int vdir, hdir, mov;
	spot_t spot;

	vdir = train.vdir;
	hdir = train.hdir;
	mov = movtype(vdir, hdir); /* the current, not changed, movement */
	doturn(&vdir, &hdir, turn);
	dstx = train.x + hdir;
	dsty = train.y + vdir;

	/* Allow wrapping around. */
	dstx = clipx(dstx);
	dsty = clipy(dsty);

	spot = field_get(dsty, dstx);
	if (spot == JUNC_UNIV) return 1;
	if (spot == JUNC_DIAG && (mov == MOV_VERT || mov == MOV_HORIZ))
		return 1;
	if (spot == JUNC_HV && (mov == MOV_ASCEND || mov == MOV_DESCEND))
		return 1;
	if (spot == RAIL_HORIZ && mov != MOV_VERT && mov != MOV_HORIZ)
		return 1;
	if (spot == RAIL_VERT && mov != MOV_HORIZ && mov != MOV_VERT)
		return 1;
	if (spot == RAIL_ASCEND && mov != MOV_DESCEND && mov != MOV_ASCEND)
		return 1;
	if (spot == RAIL_DESCEND && mov != MOV_ASCEND && mov != MOV_DESCEND)
		return 1;
	return 0;
}

static void orienttrain(void);

/* Check the train's destination -- where it'll be next move. */
static void checkdest(void)
{
	int dstx, dsty;
	int hdir = train.hdir, vdir = train.vdir;
	spot_t spot;

	dstx = train.x + hdir;
	dsty = train.y + vdir;

	/* Allow wrapping around. */
	dstx = clipx(dstx);
	dsty = clipy(dsty);

	spot = field_get(dsty, dstx);
	if (spot == TRAIN_REFLECT && !train.justreflected)
	{
		train.vdir *= -1;
		train.hdir *= -1;
		train.justreflected = 1;
		sound_play(SMP_REFLECT);
		orienttrain();
	}
	else if (spot != TRAIN_TELE
		&& !(spot == RAIL_HORIZ && hdir != NONE)
		&& !(spot == RAIL_VERT && vdir != NONE)
		&& !(spot == RAIL_ASCEND &&
			(hdir == NONE || vdir == NONE
			|| (vdir == UP && hdir == RIGHT)
			|| (vdir == DOWN && hdir == LEFT)))
		&& !(spot == RAIL_DESCEND &&
			(hdir == NONE || vdir == NONE
			|| (vdir == UP && hdir == LEFT)
			|| (vdir == DOWN && hdir == RIGHT)))
		&& !(spot == JUNC_UNIV)
		&& !(spot == JUNC_HV)
		&& !(spot == JUNC_DIAG)
		&& !(spot == JUNC_Y_UP &&
			((vdir == DOWN && hdir != NONE)
			|| (vdir == UP && hdir == NONE)))
		&& !(spot == JUNC_Y_DOWN &&
			((vdir == UP && hdir != NONE)
			|| (vdir == DOWN && hdir == NONE)))
		&& !(spot == JUNC_Y_LEFT &&
			((hdir == RIGHT && vdir != NONE)
			|| (hdir == LEFT && vdir == NONE)))
		&& !(spot == JUNC_Y_RIGHT &&
			((hdir == LEFT && vdir != NONE)
			|| (hdir == RIGHT && vdir == NONE))))
	{
		int found = -1;
		/* No primary connector, look for a secondary one. */
		if (train.turn == TURN_LEFT)
		{
			if (secondary(TURN_LEFT)) found = TURN_LEFT;
			else if (secondary(TURN_RIGHT)) found = TURN_RIGHT;
		}
		else
		{
			if (secondary(TURN_RIGHT)) found = TURN_RIGHT;
			else if (secondary(TURN_LEFT)) found = TURN_LEFT;
		}
		if (found != -1)
			doturn(&train.vdir, &train.hdir, found);
	}
}

static void teleporttrain(void)
{
	int y, x;

	y = train.y;
	x = train.x;

	/* If there is only one teleporter, we'll come back to it. */
	for (;;)
	{
		x = clipx(x + 1);
		if (train.x == x) y = clipy(y - 1);
		if (field_get(y, x) == TRAIN_TELE) break;
	}

	/* If we moved at all, play the teleport sound. */
	if (train.y != y || train.x != x)
		sound_play(SMP_TELEPORT);

	train.y = y;
	train.x = x;
}

/* Orient the train, decide where to next. Train may end up crashing. */
static void orienttrain(void)
{
	int vdir, hdir, turn;

	hdir = train.hdir;
	vdir = train.vdir;
	turn = train.turn;

	switch(field_get(train.y, train.x))
	{
		case RAIL_HORIZ: /* - */
		{
			vdir = NONE;
			if (hdir == NONE) CRASH;
			break;
		}
		case RAIL_VERT: /* | */
		{
			hdir = NONE;
			if (vdir == NONE) CRASH;
			break;
		}
		case RAIL_ASCEND: /* / */
		{
			if (hdir == RIGHT || vdir == UP)
			{
				if (hdir == LEFT || vdir == DOWN) CRASH;
				hdir = RIGHT;
				vdir = UP;
			}
			else
			{
				hdir = LEFT;
				vdir = DOWN;
			}
			break;
		}
		case RAIL_DESCEND: /* \ */
		{
			if (hdir == LEFT || vdir == UP)
			{
				if (hdir == RIGHT || vdir == DOWN) CRASH;
				hdir = LEFT;
				vdir = UP;
			}
			else
			{
				hdir = RIGHT;
				vdir = DOWN;
			}
			break;
		}
		case JUNC_UNIV: /* * */
			break; /* Universal junction, just keep on going */
		case JUNC_HV: /* + */
		{
			/*
			 * This junction can be entered from any direction.
			 * The turn value is taken into account here.
			 */
			if (vdir != NONE && hdir != NONE)
			{
				if (vdir == UP && hdir == LEFT
					&& turn == TURN_LEFT) vdir = NONE;
				else if (vdir == UP && hdir == LEFT
					&& turn == TURN_RIGHT) hdir = NONE;
				else if (vdir == UP && hdir == RIGHT
					&& turn == TURN_LEFT) hdir = NONE;
				else if (vdir == UP && hdir == RIGHT
					&& turn == TURN_RIGHT) vdir = NONE;
				else if (vdir == DOWN && hdir == LEFT
					&& turn == TURN_LEFT) hdir = NONE;
				else if (vdir == DOWN && hdir == LEFT
					&& turn == TURN_RIGHT) vdir = NONE;
				else if (vdir == DOWN && hdir == RIGHT
					&& turn == TURN_LEFT) vdir = NONE;
				else if (vdir == DOWN && hdir == RIGHT
					&& turn == TURN_RIGHT) hdir = NONE;
			}
			break;
		}
		case JUNC_DIAG: /* x */
		{
			/*
			 * This junction can be entered from any direction.
			 * The turn value is taken into account here.
			 */
			if (vdir == NONE || hdir == NONE)
			{
				if (vdir == UP && turn == TURN_LEFT)
					hdir = LEFT;
				else if (vdir == UP && turn == TURN_RIGHT)
					hdir = RIGHT;
				else if (vdir == DOWN && turn == TURN_LEFT)
					hdir = RIGHT;
				else if (vdir == DOWN && turn == TURN_RIGHT)
					hdir = LEFT;
				else if (hdir == LEFT && turn == TURN_LEFT)
					vdir = DOWN;
				else if (hdir == LEFT && turn == TURN_RIGHT)
					vdir = UP;
				else if (hdir == RIGHT && turn == TURN_LEFT)
					vdir = UP;
				else if (hdir == RIGHT && turn == TURN_RIGHT)
					vdir = DOWN;
			}
			break;
		}
		case JUNC_Y_UP: /* v */
		{
			if (vdir == DOWN && hdir != NONE) hdir = NONE;
			else if (!(vdir == UP && hdir == NONE)) CRASH;
			else if (turn == TURN_LEFT) hdir = LEFT;
			else hdir = RIGHT;
			break;
		}
		case JUNC_Y_DOWN: /* ^ */
		{
			if (vdir == UP && hdir != NONE) hdir = NONE;
			else if (!(vdir == DOWN && hdir == NONE)) CRASH;
			else if (turn == TURN_LEFT) hdir = RIGHT;
			else hdir = LEFT;
			break;
		}
		case JUNC_Y_LEFT: /* > */
		{
			if (hdir == RIGHT && vdir != NONE) vdir = NONE;
			else if (!(hdir == LEFT && vdir == NONE)) CRASH;
			else if (turn == TURN_LEFT) vdir = DOWN;
			else vdir = UP;
			break;
		}
		case JUNC_Y_RIGHT: /* < */
		{
			if (hdir == LEFT && vdir != NONE) vdir = NONE;
			else if (!(hdir == RIGHT && vdir == NONE)) CRASH;
			else if (turn == TURN_LEFT) vdir = UP;
			else vdir = DOWN;
			break;
		}
		case TRAIN_START: /* $ */
		case EMPTY:
		{
			CRASH;
			break;
		}
		case TRAIN_TELE: /* # */
		{
			/* Like an HV junction (+). Copied and pasted: */
			if (vdir != NONE && hdir != NONE)
			{
				if (vdir == UP && hdir == LEFT
					&& turn == TURN_LEFT) vdir = NONE;
				else if (vdir == UP && hdir == LEFT
					&& turn == TURN_RIGHT) hdir = NONE;
				else if (vdir == UP && hdir == RIGHT
					&& turn == TURN_LEFT) hdir = NONE;
				else if (vdir == UP && hdir == RIGHT
					&& turn == TURN_RIGHT) vdir = NONE;
				else if (vdir == DOWN && hdir == LEFT
					&& turn == TURN_LEFT) hdir = NONE;
				else if (vdir == DOWN && hdir == LEFT
					&& turn == TURN_RIGHT) vdir = NONE;
				else if (vdir == DOWN && hdir == RIGHT
					&& turn == TURN_LEFT) vdir = NONE;
				else if (vdir == DOWN && hdir == RIGHT
					&& turn == TURN_RIGHT) hdir = NONE;
			}

			/* Teleport, unless that's how you got here. */
			if (!train.justteleported)
			{
				teleporttrain();
				train.justteleported = 1;
				orienttrain();
				return;
			}
			break;
		}
		case TRAIN_REFLECT: /* @ */
			assert(0);
	}

	train.hdir = hdir;
	train.vdir = vdir;
	train.turn = turn;
	train.justteleported = 0;

	checkdest();

	train.justreflected = 0;
}

enum { NOBULLET, BUL_LEFT, BUL_RIGHT, BUL_DOWN, BUL_UP };

static void firebullet(int vdir, int hdir)
{
	int x, y;
	if (!train.canfire) return;
	y = clipy(train.y + vdir
		+ (train.steps >= STEPS_PER_MOVE/2 ? train.vdir : 0));
	x = clipx(train.x + hdir
		+ (train.steps >= STEPS_PER_MOVE/2 ? train.hdir : 0));
	spawn_bullet(y, x, vdir, hdir);
	if (!possiblyremovebullet(numbullets-1))
		sound_play(SMP_SHOOT);
	train.canfire = 0;
}

static void throwbullet(int direction)
{
	int bvdir = NONE, bhdir = NONE;
	switch (direction)
	{
		case BUL_LEFT: bhdir = LEFT; break;
		case BUL_RIGHT: bhdir = RIGHT; break;
		case BUL_UP: bvdir = UP; break;
		case BUL_DOWN: bvdir = DOWN; break;
	}
	firebullet(bvdir, bhdir);
}

static void movetrain(void)
{
	if (!train.inuse) return;

	/*
	 * Hack: The train should never crash, but some level segments
	 * might have mistakes. In case we missed one, if the train
	 * crashes, respawn it in its starting location.
	 */
	if (train.crashed)
	{
		train.x = train.spawnx;
		train.y = train.spawny;
		train.vdir = DOWN;
		train.hdir = RIGHT;
		train.crashed = 0;
		train.steps = 0;
		train.canfire = 1;
	}

	if (++train.steps == STEPS_PER_MOVE)
	{
		train.steps = 0;
		train.x = clipx(train.x + train.hdir);
		train.y = clipy(train.y + train.vdir);
		orienttrain();
	}
	else if (train.steps == STEPS_PER_MOVE/2)
		train.canfire = 1;
}

static void explodetrain(void)
{
	if (train.inuse && train.crashed)
	{
		if (train.vdir || train.hdir)
		{
			train.vdir = train.hdir = NONE;
			train.turn = TURN_LEFT;
		}
		else if (train.turn == TURN_LEFT)
			train.turn = TURN_RIGHT;
	}
}

static void turnmonster(int mt)
{
	int newdir, oldhdir, oldvdir;

	oldhdir = monsters[mt].hdir;
	oldvdir = monsters[mt].vdir;

	do
	{
		newdir = rnd(4);
		if (newdir&1)
		{
			monsters[mt].vdir = NONE;
			monsters[mt].hdir = (newdir&2) ? LEFT : RIGHT;
		}
		else
		{
			monsters[mt].hdir = NONE;
			monsters[mt].vdir = (newdir&2) ? UP : DOWN;
		}
	} while (monsters[mt].hdir == oldhdir
		&& monsters[mt].vdir == oldvdir);

	monsters[mt].steps = -MONSTER_TURN_WAIT;
}

static void checkmonstersanity(int mt)
{
	assert(monsters[mt].hdir >= -1 && monsters[mt].hdir <= 1);
	assert(monsters[mt].vdir >= -1 && monsters[mt].vdir <= 1);
	assert(monsters[mt].steps >= -MONSTER_WAIT);
	assert(monsters[mt].steps < STEPS_PER_MOVE);
	assert(monsters[mt].x == clipx(monsters[mt].x));
	assert(monsters[mt].y == clipx(monsters[mt].y));
}

static void checkbulletsanity(int bt)
{
	assert(bullets[bt].hdir >= -1 && bullets[bt].hdir <= 1);
	assert(bullets[bt].vdir >= -1 && bullets[bt].vdir <= 1);
	assert(bullets[bt].steps >= -MONSTER_WAIT);
	assert(bullets[bt].steps < STEPS_PER_MOVE);
	assert(bullets[bt].x == clipx(bullets[bt].x));
	assert(bullets[bt].y == clipx(bullets[bt].y));
}

static void movemonster(int mt)
{
	checkmonstersanity(mt);
	if (++monsters[mt].steps == STEPS_PER_MOVE)
	{
		monsters[mt].steps = -MONSTER_WAIT;
		monsters[mt].x = clipx(monsters[mt].x + monsters[mt].hdir);
		monsters[mt].y = clipy(monsters[mt].y + monsters[mt].vdir);
	}
	else if (monsters[mt].steps == 0)
	{
		int mdsty = monsterdsty(mt), mdstx = monsterdstx(mt);
		if (rnd(4) == 0
			|| field_get(mdsty, mdstx) != EMPTY
			|| monsterinspot(mdsty, mdstx, mt) != -1)
		{
			turnmonster(mt);
		}
	}
	checkmonstersanity(mt);
}

static void movemonsters(void)
{
	int i;
	for (i = 0; i < nummonsters; i++)
	{
		if (monsters[i].alive < 0)
		{
			monsters[i].alive++;
			if (monsters[i].alive == 0 && numlivingmonsters == 0)
				newlevel();
		}
		else if (monsters[i].alive)
			movemonster(i);
	}
}

static void movebullet(int bt)
{
	checkbulletsanity(bt);
	if (++bullets[bt].steps == STEPS_PER_MOVE)
	{
		bullets[bt].steps = 0;
		bullets[bt].x = clipx(bullets[bt].x + bullets[bt].hdir);
		bullets[bt].y = clipy(bullets[bt].y + bullets[bt].vdir);
		if (possiblyremovebullet(bt))
			return;
	}
	checkbulletsanity(bt);
}

static void movebullets(void)
{
	int i;
	for (i = 0; i < numbullets; i++)
		if (bullets[i].inuse)
			movebullet(i);
}

/************************************************************/

static void handlekey(int keynum)
{
	if (titlescreen)
	{
		if (keynum == KEY_ESC) gameover = 1;
		else if (keynum == KEY_SPACE) newlevel();
		return;
	}

	switch (keynum)
	{
		case KEY_LEFT:
			train.turn = TURN_LEFT;
			break;
		case KEY_RIGHT:
			train.turn = TURN_RIGHT;
			break;
		case KEY_6_PAD:
		case KEY_D: /* WASD */
		case KEY_K: /* UHJK */
			throwbullet(BUL_RIGHT);
			break;
		case KEY_4_PAD:
		case KEY_A: /* WASD */
		case KEY_H: /* UHJK */
			throwbullet(BUL_LEFT);
			break;
		case KEY_2_PAD:
		case KEY_S: /* WASD */
		case KEY_J: /* UHJK */
			throwbullet(BUL_DOWN);
			break;
		case KEY_8_PAD:
		case KEY_W: /* WASD */
		case KEY_U: /* UHJK */
			throwbullet(BUL_UP);
			break;
		case KEY_SPACE:
			/*
			 * Pausing takes effect once logic completes.
			 * If the TIME OVER banner is up, don't pause,
			 * instead restart.
			 */
			if (train.gameovercount) restartlevel();
			else pause_game();
			break;
		case KEY_ESC:
			titlescreen = 1;
			reset_colors();
			sound_playtitlemusic();
			break;
		case KEY_ENTER:
			if (key_shifts & KB_ALT_FLAG)
				toggle_fullscreen();
			break;
		default:
			/* Whatever, man. I can't heeear you! */
			break;
	}
}

static void readkeys(void)
{
	while (keypressed())
		handlekey(readkey()>>8);
}

/* How long before you restart, after running out of time. */
#define GAMEOVERSECS 10

int game_run_logic_frame(void)
{
	int changed = 0; /* did anything actually change here? */

	if (titlescreen)
	{
		readkeys();
		return 0;
	}
	readkeys();
	tics++;
	if (tics % TICS_PER_STEP == 0)
	{
		movetrain();
		explodetrain();
		changed = 1;
	}
	if (tics % TICS_PER_MONSTER_STEP == 0)
	{
		movemonsters();
		changed = 1;
	}
	if (tics % TICS_PER_BULLET_STEP == 0)
	{
		movebullets();
		changed = 1;
	}
	if (tics == TICS_PER_STEP * TICS_PER_MONSTER_STEP
		* TICS_PER_BULLET_STEP) tics = 0;

	if (train.gameovercount)
	{
		train.gameovercount--;
		if (train.gameovercount == 0) restartlevel();
	}
	else if (--timetics == 0)
	{
		/* Bummer, man. You ran out of time. */
		sound_play(SMP_TIMEUP);
		train.crashed = 1;
		explodetrain();
		train.gameovercount = GAMEOVERSECS * TICS;

		/*
		 * This function name is misleading. It actually just
		 * stops the music, not the time-up sound effect.
		 */
		sound_stop();
	}

	/* If we're returning to the title screen, it needs to be drawn. */
	if (titlescreen) changed = 1;

	return changed;
}

/************************************************************/

/*
 * Sure, the levels are FIELD_WIDTH by FIELD_HEIGHT. But how
 * much is actually on-screen at a time?
 */
#define VIEW_WIDTH  15
#define VIEW_HEIGHT 15

#define SCRN_TOP ((SCRHEIGHT - (VIEW_HEIGHT*FONT_HEIGHT))/2)
#define SCRN_LEFT ((SCRWIDTH - (VIEW_WIDTH*FONT_WIDTH))/2)
#define TRAIN_TO_NUMERAL_COLOR 72 /* XXX unused */
#define NUMERAL_COLOR 8 /* dark gray */

static char spot2char(spot_t spot)
{
	const char *screenchars = "-|/\\*+xv^><$#@.";
	return screenchars[spot];
}

#define STATUS_H_MARGIN 2
#define STATUS_W (SCRN_LEFT - STATUS_H_MARGIN*2)

/* Left and right of status1 (left side) and status2 (right side). */
#define STATUS1_LEFT STATUS_H_MARGIN
#define STATUS1_RIGHT (STATUS_H_MARGIN + STATUS_W - 1)
#define STATUS1_CENTER ((STATUS1_LEFT+STATUS1_RIGHT)/2)
#define STATUS2_LEFT (SCRWIDTH - SCRN_LEFT + STATUS_H_MARGIN)
#define STATUS2_RIGHT (SCRWIDTH - STATUS_H_MARGIN)
#define STATUS2_CENTER ((STATUS2_LEFT+STATUS2_RIGHT)/2)

/* The small font's size. */
#define SFONT_WIDTH  8
#define SFONT_HEIGHT 8

#define VSPACE 2

#define LEGEND_MEDIAN 24
#define LEGEND_HEIGHT (3*FONT_HEIGHT + 11*VSPACE \
                    + 11*SFONT_HEIGHT + 2*LEGEND_MEDIAN)
#define LEGEND_TOP ((SCRHEIGHT-LEGEND_HEIGHT)/2)

/* Draw the legend, or "help screen." */
static void drawlegend(BITMAP *target)
{
	int y, x;

	y = LEGEND_TOP;
	x = STATUS1_CENTER;

	draw_sprite(target, image_train(DOWN, RIGHT, TURN_RIGHT),
		x - (FONT_WIDTH/2), y);
	y += FONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Train", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Turn with", x, y, status_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "arrow", x, y, status_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "keys.", x, y, status_color, -1);
	y += SFONT_HEIGHT + LEGEND_MEDIAN;

	draw_sprite(target, image_bullet, x - (FONT_WIDTH/2), y);
	y += FONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Acorns", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Fire them", x, y, status_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "with WASD", x, y, status_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "keys.", x, y, status_color, -1);
	y += SFONT_HEIGHT + LEGEND_MEDIAN;

	draw_sprite(target, image_enemy(NONE, RIGHT, 0), x - (FONT_WIDTH/2), y);
	y += FONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Squirrels", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Kill", x, y, status_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "them all.", x, y, status_color, -1);
	y += SFONT_HEIGHT;
}

static char *gettimerstring(void)
{
	static char timerstring[] = "x:xx";
	unsigned int timesecs = timetics / TICS;

	/* This drawer only handles up to 9:59, natch. */
	assert(timesecs < 600);

	timerstring[0] = '0' + timesecs/60;
	timerstring[2] = '0' + timesecs%60/10;
	timerstring[3] = '0' + timesecs%10;
	return timerstring;
}

#define RSTATUS_MEDIAN1 20
#define RSTATUS_MEDIAN2 (SCRHEIGHT/2)
#define RSTATUS_HEIGHT (7*SFONT_HEIGHT + 4*VSPACE + RSTATUS_MEDIAN1 \
                        + RSTATUS_MEDIAN2)
#define RSTATUS_TOP ((SCRHEIGHT-RSTATUS_HEIGHT)/2)

/* Draw the right status bar: number of monsters living; timer. */
static void drawstatus(BITMAP *target)
{
	char monstr[STATUS_W/FONT_WIDTH + 1];
	int y, x;

	y = RSTATUS_TOP;
	x = STATUS2_CENTER;

	snprintf(monstr, sizeof monstr, "%d", numlivingmonsters);

	textout_centre_ex(target, font, "Squirrels", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Left", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, monstr, x, y, status_color, -1);
	y += SFONT_HEIGHT + RSTATUS_MEDIAN1;

	textout_centre_ex(target, font, "Time Left", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, gettimerstring(), x, y,
		status_color, -1);
	y += SFONT_HEIGHT + RSTATUS_MEDIAN2;

	textout_centre_ex(target, font, "Rail", x, y, status_em_color, -1);
	y += SFONT_HEIGHT + VSPACE;
	textout_centre_ex(target, font, "Blaster", x, y, status_em_color, -1);
	y += SFONT_HEIGHT;
}

/* Draw the legend and status info. */
static void drawstatusnonsense(BITMAP *target)
{
	drawlegend(target);
	drawstatus(target);
}

/************************************************************/

/* Title screen drawing */

/* Draw a string in the big magnified "font." */
static void bigtext(BITMAP *target, int x, int y, const char *s,
	int color, int bg)
{
	x -= FONT_WIDTH * strlen(s) / 2;
	while (*s != '\0')
	{
		drawchar(target, x, y, *s, color, bg);
		x += FONT_WIDTH;
		s++;
	}
}

#define TITLE_COLOR 10 /* bright green */
#define INFO_COLOR 7 /* gray */
#define SPACE_COLOR 11 /* bright cyan */
#define TITLE_HEIGHT (FONT_HEIGHT*10 + SFONT_HEIGHT)
#define TITLE_TOP ((SCRHEIGHT-TITLE_HEIGHT)/2)

static void drawtitle(BITMAP *target)
{
	int y = TITLE_TOP, x = SCRWIDTH/2;

	clear_bitmap(target);

	bigtext(target, x, y, "Rail Blaster", TITLE_COLOR, -1);
	y += 3*FONT_HEIGHT;

	bigtext(target, x, y, "-|/\\    Tracks     ", INFO_COLOR, -1);
	y += FONT_HEIGHT;
	bigtext(target, x, y, "*+x     Junctions  ", INFO_COLOR, -1);
	y += FONT_HEIGHT;
	bigtext(target, x, y, "v^><    Y-junctions", INFO_COLOR, -1);
	y += FONT_HEIGHT;
	bigtext(target, x, y, "@       Reflectors ", INFO_COLOR, -1);
	y += FONT_HEIGHT;
	bigtext(target, x, y, "#       Teleporters", INFO_COLOR, -1);
	y += 3*FONT_HEIGHT;

	textout_centre_ex(target, font, "   P R E S S   S P A C E   ",
		x, y, SPACE_COLOR, -1);
}

/************************************************************/

/*
 * spoty and spotx are in tiles in the level. But they may be off a side.
 * viewy and viewx are in pixels relative to the top left of the screen.
 */
static void drawspot(BITMAP *target, int spoty, int spotx, int viewy, int viewx)
{
	spoty = clipy(spoty);
	spotx = clipx(spotx);
	drawchar(target, viewx, viewy,
		spot2char(field_get(spoty, spotx)), view_color, view_bg);
}

/* The train is always, always in the center of the screen. */
#define TRAINX ((SCRWIDTH  - FONT_WIDTH)  / 2)
#define TRAINY ((SCRHEIGHT - FONT_HEIGHT) / 2)

void game_draw_frame(BITMAP *target)
{
	int y, x, i;
	int trainx, trainy;
	int basespoty, basespotx;
	int basepixely, basepixelx;
	int numspotstodrawy = VIEW_HEIGHT, numspotstodrawx = VIEW_WIDTH;

	if (titlescreen)
	{
		drawtitle(target);
		return;
	}

	clear_to_color(target, status_bg);
	set_clip_rect(target, SCRN_LEFT, SCRN_TOP,
		SCRN_LEFT + VIEW_WIDTH*FONT_WIDTH - 1,
		SCRN_TOP + VIEW_HEIGHT*FONT_HEIGHT - 1);

	trainx = (SCRWIDTH  - FONT_WIDTH)  / 2;
	trainy = (SCRHEIGHT - FONT_HEIGHT) / 2;

	/* The spot that'll show in the top left of the screen */
	basespotx = train.x - VIEW_WIDTH/2;
	basespoty = train.y - VIEW_WIDTH/2;

	/* The pixels where this spot will show */
	basepixelx = SCRN_LEFT;
	basepixely = SCRN_TOP;

	/* Reconcile the base variables with train's sub-spot movement */
	if (train.steps != 0)
	{
		if (train.vdir != NONE) numspotstodrawy++;
		if (train.hdir != NONE) numspotstodrawx++;
		basepixelx -= train.hdir * train.steps * FONT_WIDTH
			/ STEPS_PER_MOVE;
		basepixely -= train.vdir * train.steps * FONT_HEIGHT
			/ STEPS_PER_MOVE;
		if (basepixelx > SCRN_LEFT)
		{
			basespotx--;
			basepixelx -= FONT_WIDTH;
		}
		if (basepixely > SCRN_TOP)
		{
			basespoty--;
			basepixely -= FONT_HEIGHT;
		}
	}

	/* Clip the base spot -- drawing enemies later on needs this. */
	basespotx = clipx(basespotx);
	basespoty = clipy(basespoty);

	for (y = basespoty; y < basespoty + numspotstodrawy; y++)
		for (x = basespotx; x < basespotx + numspotstodrawx; x++)
		{
			drawspot(target, y, x,
				(y-basespoty)*FONT_HEIGHT + basepixely,
				(x-basespotx)*FONT_WIDTH  + basepixelx);
		}

	if (train.inuse)
	{
		masked_blit(image_train(train.vdir, train.hdir, train.turn),
			target,
			0, 0,
			trainx, trainy,
			FONT_WIDTH, FONT_HEIGHT);
	}

	for (i = 0; i < nummonsters; i++)
	{
		int monx, mony;
		int imgstatus;

		if (!monsters[i].alive) continue;

		checkmonstersanity(i);

		mony = clipy(monsters[i].y - basespoty + 1) - 1;
		monx = clipx(monsters[i].x - basespotx + 1) - 1;

		/*
		 * Ensure that monsters just entering the view area
		 * from the top or left are visible.
		 */
		if (monsters[i].steps > 0)
		{
			if (mony == FIELD_HEIGHT-1
				&& monsters[i].vdir == DOWN)
			{
				mony -= FIELD_HEIGHT;
			}
			if (monx == FIELD_WIDTH-1
				&& monsters[i].hdir == RIGHT)
			{
				monx -= FIELD_WIDTH;
			}
		}

		mony *= FONT_HEIGHT;
		monx *= FONT_WIDTH;

		mony += basepixely;
		monx += basepixelx;

		if (monsters[i].steps > 0)
		{
			monx += monsters[i].hdir * monsters[i].steps
				* FONT_WIDTH / STEPS_PER_MOVE;
			mony += monsters[i].vdir * monsters[i].steps
				* FONT_HEIGHT / STEPS_PER_MOVE;
		}

		if (monsters[i].alive < 0) imgstatus = 2; /* collapsing */
		else if (monsters[i].steps > 0) imgstatus = 1; /* moving */
		else imgstatus = 0; /* still */

		/*
		 * Yes, I blit every monster, even though it might be
		 * totally offscreen. The view area is clipped, so
		 * it's ok. No blitting is actually getting done if
		 * the enemy is offscreen.
		 */
		masked_blit(image_enemy(monsters[i].vdir, monsters[i].hdir,
			imgstatus),
			target,
			0, 0,
			monx, mony,
			FONT_WIDTH, FONT_HEIGHT);		
	}

	/* XXX mostly a copy and paste from above */
	for (i = 0; i < numbullets; i++)
	{
		int bulx, buly;
		if (!bullets[i].inuse) continue;

		checkbulletsanity(i);

		buly = clipy(bullets[i].y - basespoty + 1) - 1;
		bulx = clipx(bullets[i].x - basespotx + 1) - 1;

		/*
		 * Ensure that bullets just entering the view area
		 * from the top or left are visible.
		 */
		if (bullets[i].steps > 0)
		{
			if (buly == FIELD_HEIGHT-1
				&& bullets[i].vdir == DOWN)
			{
				buly -= FIELD_HEIGHT;
			}
			if (bulx == FIELD_WIDTH-1
				&& bullets[i].hdir == RIGHT)
			{
				bulx -= FIELD_WIDTH;
			}
		}

		buly *= FONT_HEIGHT;
		bulx *= FONT_WIDTH;

		buly += basepixely;
		bulx += basepixelx;

		if (bullets[i].steps > 0)
		{
			bulx += bullets[i].hdir * bullets[i].steps
				* FONT_WIDTH / STEPS_PER_MOVE;
			buly += bullets[i].vdir * bullets[i].steps
				* FONT_HEIGHT / STEPS_PER_MOVE;
		}

		/*
		 * The same excuse above for blitting every monster
		 * applies here too.
		 */
		masked_blit(image_bullet,
			target, /* no pun intended */
			0, 0,
			bulx, buly,
			FONT_WIDTH, FONT_HEIGHT);		
	}

	set_clip_rect(target, 0, 0, SCRWIDTH-1, SCRHEIGHT-1);

	drawstatusnonsense(target);

	if (train.gameovercount)
	{
		/* Print a "TIME OVER" banner. */
		textout_centre_ex(target, font, "   T I M E   O V E R   ",
			SCRWIDTH/2, SCRHEIGHT/2, status_em_color, -1);
	}
}
