#ifdef NO_MUSIC
void sound_init(void) { }
void sound_stop(void) { }
void sound_start(void) { }
void sound_update(void) { }
void sound_play(int sample) { }
void sound_nextmusic(void) { }
void sound_playtitlemusic(void) { }
#else

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <allegro.h>
#include "modplug.h"
#include "main.h"
#include "config.h"
#include "sound.h"

#define NUMMUSICS 2 /* If you change this, change sound_init() below. */
static ModPlugFile *mods[NUMMUSICS+1]; /* Hack: mods[NUMMUSICS] is title. */
static int playing = NUMMUSICS; /* Title music. */
static int soundinited; /* If the sound has initialized or not. */
static int soundstopped; /* Temporarily stopped? (pause/gameover screens) */

/*
 * Playing is initialized to 1 as a hack, so that 0 actually plays first.
 * The music gets changed when a new level starts, you see.
 */

static void *readfile(const char *filename, int *len)
{
	char *buf = NULL;
	int nbuf = 0, sbuf = 0;
	int c;
	FILE *fp = fopen(filename, "rb");

	if (fp == NULL) abort_with_error("File %s: cannot open", filename);

	while ((c = getc(fp)) != EOF)
	{
		XPND(buf, nbuf, sbuf);
		buf[nbuf++] = c;
	}
	fclose(fp);
	*len = nbuf;
	return buf;
}

#define MUSIC_BITS 8 /* Dammit! If I do 16 it gets all distorted... */
#define MUSIC_MODPLUG_BITS 16
#define MUSIC_STEREO TRUE
#define MUSIC_FREQ 44100
#define MUSIC_VOL 255
#define MUSIC_PAN 128
#define MUSIC_CHANNELS (MUSIC_STEREO ? 2 : 1)
#define MUSIC_BYTES (MUSIC_BITS/8)

/*
 * Default settings.
 *
 * Don't use noise reduction; Modplug Tracker's help advises against it in
 * fact, as it's just a naive lowpass filter.
 */
#define MP2WAV_DEF_FLAGS          MODPLUG_ENABLE_OVERSAMPLING \
                                | MODPLUG_ENABLE_SURROUND
#define MP2WAV_DEF_CHANNELS       (MUSIC_STEREO ? 2 : 1)
#define MP2WAV_DEF_BITS           MUSIC_MODPLUG_BITS
#define MP2WAV_DEF_SAMPLE_RATE    MUSIC_FREQ
#define MP2WAV_DEF_RESAMPLE       MODPLUG_RESAMPLE_LINEAR
#define MP2WAV_DEF_XBASS_AMOUNT   50 /* 0 to 100 */
#define MP2WAV_DEF_XBASS_RANGE    55 /* cutoff in Hz */
#define MP2WAV_DEF_SURROUND_DEPTH 75 /* 0 to 100 */
#define MP2WAV_DEF_SURROUND_DELAY 20 /* usually 5 to 40 ms */
#define MP2WAV_DEF_LOOP_COUNT     -1 /* loop forever */

static ModPlug_Settings settings =
{
	MP2WAV_DEF_FLAGS,
	MP2WAV_DEF_CHANNELS,
	MP2WAV_DEF_BITS,
	MP2WAV_DEF_SAMPLE_RATE,
	MP2WAV_DEF_RESAMPLE,
	0, 0, /* reverb settings, unused for now */
	MP2WAV_DEF_XBASS_AMOUNT,
	MP2WAV_DEF_XBASS_RANGE,
	MP2WAV_DEF_SURROUND_DEPTH,
	MP2WAV_DEF_SURROUND_DELAY,
	MP2WAV_DEF_LOOP_COUNT
};

static AUDIOSTREAM *stream;
#define SOUND_BUF_SIZE 16384 /* This is in sample frames */
#define SOUND_BUF_SAMPLES (SOUND_BUF_SIZE * MUSIC_CHANNELS)
#define SOUND_BUF_BYTES (SOUND_BUF_SIZE * MUSIC_BYTES * MUSIC_CHANNELS)
static short my16bitbuffer[SOUND_BUF_SAMPLES];

#define NUM_VOICES 4 /* Ought to be enough for anybody... */
static void loadwavs(void); /* Defined below, called from here */

static ModPlugFile *loadmod(const char *filename)
{
	void *filedata;
	int len;
	ModPlugFile *ret;

	filedata = readfile(filename, &len);
	ret = ModPlug_Load(filedata, len);
	if (ret == NULL)
		abort_with_error("Cannot load music file %s", filename);
	free(filedata);
	return ret;
}

void sound_init(void)
{
	assert(soundinited == 0);

	/* Setup on the modplug side */
	ModPlug_SetSettings(&settings);
	mods[0] = loadmod(SHAREPATH "Music-1.it");
	mods[1] = loadmod(SHAREPATH "Music-2.it");
	mods[2] = loadmod(SHAREPATH "Music-0.it");

	/* Load the wav files */
	loadwavs();

	/* Setup on the Allegro side */
	reserve_voices(NUM_VOICES, -1);
	if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) != 0)
		abort_with_error("Cannot initialize sound: %s", allegro_error);
	stream = play_audio_stream(SOUND_BUF_SIZE,
		MUSIC_BITS, MUSIC_STEREO, MUSIC_FREQ, MUSIC_VOL, MUSIC_PAN);
	if (stream == NULL)
	{
		abort_with_error("Error creating audio stream: %s",
			allegro_error);
	}

	soundinited = 1;
}

#define SAT_GRAD 0.4f
#define SAT_GRADINV (1.0f / SAT_GRAD)
#define SAMP_RANGE 32767.0f
#define SAT_RANGE (SAMP_RANGE * (1.0f / ((SAT_GRAD + 1.0f) / 2.0f)))
static int saturate(double sample)
{
	sample /= SAT_RANGE;
	if (sample > 0.0f)
	{
		if (sample > SAT_GRAD)
		{
			sample = SAT_GRAD + (sample-SAT_GRAD)
				/ (1+((sample-SAT_GRAD)/(1-SAT_GRAD))
				   * ((sample-SAT_GRAD)/(1-SAT_GRAD)));
		}
		if (sample > 1.0f)
			sample = (SAT_GRAD + 1.0f) / 2;
	}
	else
	{
		sample = -sample;
		if (sample > SAT_GRAD)
		{
			sample = SAT_GRAD + (sample-SAT_GRAD)
				/ (1+((sample-SAT_GRAD)/(1-SAT_GRAD))
				   * ((sample-SAT_GRAD)/(1-SAT_GRAD)));
		}
		if (sample > 1.0f)
			sample = (SAT_GRAD + 1.0f) / 2;
		sample = -sample;
	}
	sample *= SAT_RANGE;
	/* XXX pointless? */
	if (sample > SAMP_RANGE) sample = SAMP_RANGE;
	else if (sample < -SAMP_RANGE) sample = -SAMP_RANGE;
	return (int)sample;
}

void sound_restartmusic(void)
{
	ModPlug_Seek(mods[playing], 0); /* Back to the beginning. */
}

#define MODPLUG_BUF_BYTES (SOUND_BUF_BYTES * MUSIC_MODPLUG_BITS / MUSIC_BITS)

/* The output from modplug is really quiet. This is a multiplier for it. */
#define AMP 6.0f

void sound_update(void)
{
	unsigned char *p;
	int nbytes, i;

	if (soundstopped) return;

	p = get_audio_stream_buffer(stream);
	if (p == NULL) return; /* Nothing to do at the moment */

	nbytes = ModPlug_Read(mods[playing], my16bitbuffer, MODPLUG_BUF_BYTES);
	if (nbytes < MODPLUG_BUF_BYTES)
	{
		/* Mod must've ended, seek to beginning. */
		sound_restartmusic();
		nbytes += ModPlug_Read(mods[playing], &my16bitbuffer[nbytes],
			MODPLUG_BUF_BYTES-nbytes);

		/* If we STILL don't have enough, something is wrong. */
		if (nbytes < MODPLUG_BUF_BYTES)
		{
			abort_with_error("Music: Wanted %d bytes, only got %d",
				MODPLUG_BUF_BYTES, nbytes);
		}
	}

	/* Amplify the sound and stuff. */
	for (i = 0; i < SOUND_BUF_SAMPLES; i++)
	{
		p[i] = (unsigned char)
			(saturate(my16bitbuffer[i] * AMP)/256
			+ 128);
	}

	free_audio_stream_buffer(stream);
}

void sound_stop(void)
{
	soundstopped = 1;
	voice_stop(stream->voice);
}

void sound_start(void)
{
	soundstopped = 0;
	voice_start(stream->voice);
}

const char *samplefilenames[NUMSAMPLES] =
{
	SHAREPATH "Sshoot.wav",
	SHAREPATH "Skill.wav",
	SHAREPATH "Sdis.wav",
	SHAREPATH "Stele.wav",
	SHAREPATH "Stime.wav",
	SHAREPATH "Srefl.wav"
};

static SAMPLE *samples[NUMSAMPLES];

static void loadwavs(void)
{
	int i;
	for (i = 0; i < NUMSAMPLES; i++)
	{
		samples[i] = load_wav(samplefilenames[i]);
		if (samples[i] == NULL)
		{
			abort_with_error("Can't load sample %s: %s",
				samplefilenames[i], allegro_error);
		}
	}
}

void sound_play(int sample)
{
	int voiceinuse = -1;
	int newvoice;

	newvoice = play_sample(samples[sample], 255, 128, 1000, 0);
	if (newvoice == -1 && voiceinuse != -1)
	{
		/* Try to free up the last one. */
		deallocate_voice(voiceinuse);
		newvoice = play_sample(samples[sample], 255, 128, 1000, 0);
	}
	if (newvoice != -1) voiceinuse = newvoice;
}

/* Switch to the next music file, like on a new level. */
void sound_nextmusic(void)
{
	assert(soundinited);

	/* Use first song after title screen. */
	if (playing == NUMMUSICS) playing--;

	playing = (playing + 1) % NUMMUSICS;
	assert(mods[playing] != NULL);

	/* Make sure we start the new music at the beginning. */
	sound_restartmusic();
}

void sound_playtitlemusic(void)
{
	playing = NUMMUSICS;
	sound_restartmusic();
}

#endif
